Module:    Scheduler
Synopsis:  A soft realtime scheduler
Author:    Ron Franke-Polz
License:   This software is provided 'as-is', without any express or implied warranty. 
           In no event will the author be held liable for any damages arising from 
           the use of this software.
           Permission is granted to anyone to use this software for any purpose, 
           including commercial applications, and to alter it and redistribute it 
           freely, subject to the following restrictions:
           1. The origin of this software must not be misrepresented; you must not 
              claim that you wrote the original software. If you use this software in 
              a product, an acknowledgment in the product documentation would be appreciated 
              but is not required.
           2. Altered source versions must be plainly marked as such, and must not 
              be misrepresented as being the original software.
           3. This notice may not be removed or altered from any source distribution.

// ====================================================
//        classes/types defined
// ====================================================
//
//    <attribute-leaf>
//
//    <task-queue-entry>
//    <task-queue>
//    <task-queue-mgr>
//
//    note - these classes are "sealed" to allow the
//           fastest possible method dispatch on them.
// ====================================================

// ====================================================
//        Globals defined
// ====================================================
define constant $application-name :: <byte-string> = "Scheduler Browser";
define constant $application-major-version :: <byte-string> = "0";
define constant $application-minor-version :: <byte-string> = "1";

define method application-full-name () => (full-name :: <byte-string>)
  concatenate($application-name, " Version ",
              $application-major-version, ".",
              $application-minor-version)
end method application-full-name;


// ====================================================
//    <attribute-leaf>
//
//    - defines a displayable attribute for a tree-control
//    - it can have other children, which the tree-control
//      will then try to display using: name, children?, children
// ====================================================

define class <attribute-leaf> (<object>)
    slot name :: <byte-string>,
      init-keyword: name:;

    slot children :: <stretchy-vector> = #(),
      init-keyword: children:;
end class;

define method children? (a ::<attribute-leaf>)
    ~empty?(a.children)
end method;





// ====================================================
//    <task-queue-entry>
//
//    - defines an individual task that is run in a queue.
// ====================================================

define sealed class <task-queue-entry> (<object>)

  slot name :: <byte-string>,               // name of the entry - used for any reporting
    init-keyword: name:;

  slot enabled? :: <boolean> = #t,          // should this item run?
    init-keyword: enabled:;

  slot threaded? :: <boolean> = #f,         // does this entry use a seperate thread?
    init-keyword: threaded:;

  slot thread :: false-or(<thread>) = #f;   // the thread (if threaded?) running the call-back

  slot call-back :: <function>,             // a function to call - probably a curried function
    init-keyword: call-back:;               // over an object.

end class;


// ====================================================
//    <task-queue-entry>
// ====================================================
//    execution methods
// ====================================================

// execute
//
define method execute (a :: <task-queue-entry>)
    if (a.enabled?)

      if (a.threaded?)
        // create a seperate thread for the call-back
        a.thread = make(<thread>, name: a.name, function: a.call-back);
      else
        // use this thread for the call-back
        a.call-back();
      end if

    end if;
end method;


// ====================================================
//    <task-queue-entry>
// ====================================================
//    tree methods
// ====================================================

// children?
//
// - no children
//
define method children? (a :: <task-queue-entry>)
    #f;
end method;




// ====================================================
//    <task-queue>
//
//    - defines a list of tasks (function calls)
//      to be run at a rate
// ====================================================

define sealed class <task-queue> (<object>)

  slot name :: <byte-string>,               // name of the <task-queue> - used for any reporting
    init-keyword: name:;

  slot task-entries :: <stretchy-vector>    // list of <task-queue-entry>'s
    = make(<stretchy-vector>);

  slot rate :: <duration>,                  // how often this queue should run
    init-keyword: rate:;

  slot timebase-notification
       :: <timebase-notification>;

  slot timebase-thread                      // the <timebase-notification> thread for this queue
       :: false-or(<thread>) = #f;

  slot task-thread
       :: false-or(<thread>) = #f;

end class;


// initialize
//
define method initialize (a :: <task-queue>, #keys)
  a.timebase-notification := make(<timebase-notification>, name: a.name, rate: a.rate);
end method;



// ====================================================
//   <task-queue>
// ====================================================
//   task-queue thread routines
// ====================================================

// start-timebase
//
// - obviously, start up the time-base
// - this needs to be done before you "start-queue"
//
define method start-timebase (a :: <task-queue>)
  a.timebase-thread := start-notification-thread(a.timebase-notification);
end method;


// start-queue
//
// - execute the <task-queue-entry>'s
//   - create a local method that executes all of the entries everytime
//     the timebase signals.
//   - if the timebase is running,
//     create a thread that executes this local method.
// - return if we sucessfully started the queue.
//
define method start-queue (a :: <task-queue>) => (sucess? :: <boolean>);

  local method execute-entries ()
    with-timebase(a.timebase-notification)
      map(execute, a.task-entries);
    end;
  end;

  if (a.timebase-thread)
    a.task-thread := make(<thread>, name: a.name, function: execute-entries);
    #t;
  else
    #f;
  end if;

end method;


// ====================================================
//    <task-queue>
// ====================================================
//    tree methods
// ====================================================


// children
//
// - return the list of <task-entry>s
//
define method children (a :: <task-queue>)
    a.task-entries;
end method;


// children?
//
// - do we have any <task-entry>s in the list?
//
define method children? (a :: <task-queue>)
    ~empty?(a.task-entries);
end method;





// ====================================================
// <task-queue-mgr>
//
//  - a non-preemptive, multi-queue, soft realtime
//    task manager.
//  - defines a list of queues to manage, where each queue
//    has a timebase and a list of entries to execute.
//  - a table of queue names and the queues themselves is
//    managed.
// ====================================================

define sealed class <task-queue-mgr> (<object>)

  slot name :: <byte-string>,               // name of the scheduler
    init-keyword: name:;

  slot task-queues :: <stretchy-vector>     // list of <task-queue>'s
    = make(<stretchy-vector>);

  slot queue-table :: <string-table>        // a table to get queues by name
    = make(<string-table>);

  slot continue :: <boolean> = #f;          // scheduling status

end class;


// ====================================================
//    <task-queue-mgr>
// ====================================================
//    thread management methods (exported)
// ====================================================

// start-scheduler
//
// - start each of the task-queue's timebases
// - start each task-queue's execution
//
define method start-scheduler (a :: <task-queue-mgr>)

    map(start-timebase, a.task-queues);
    map(start-queue, a.task-queues);

end method;


// ====================================================
//    <task-queue-mgr>
// ====================================================
// task-queue management methods (exported)
// ====================================================

// create-queue
//
// - create a <task-queue>, put it in the task list and
//   the lookup table.
//
define method create-queue (a :: <task-queue-mgr>,
                            #key name: a-name :: <byte-string>,
                                 rate: a-rate :: <duration>)

  let new-queue = make(<task-queue>, name: a-name, rate: a-rate);

  // put the queue in the list of queue
  add!(a.task-queues, new-queue);

  // and make a look-up entry for it
  a.queue-table[a-name] := new-queue;

end method;


// schedule
//
// - place a task/function in a named queue
//   - return #t if queue found and task entered.
//   - return #f if queue not found
//
define method schedule (a :: <task-queue-mgr>,
                        #key task: task :: <function>,
                             name: name :: <byte-string>,
                            queue: a-queue-name :: <byte-string>)
                        =>(scheduled? :: <boolean>)

  // find the named <task-queue>
  let a-task-queue = a.queue-table[a-queue-name];
  if (a-task-queue)
    // create and add a <task-queue-entry> to the queue
    add!(a-task-queue.task-entries,
         make(<task-queue-entry>, name: name, call-back: task));

    // yes, we scheduled it
    #t;
  else
    // no, it wasn't scheduled
    #f;
  end if;

end method;


// ====================================================
//    <task-queue-mgr>
// ====================================================
//    tree methods
// ====================================================


// children
//
// - return the list of <task-queue>s
//
define method children (a :: <task-queue-mgr>)
    a.task-queues;
end method;


// children?
//
// - do we have any <task-queue>s in the list?
//
define method children? (a :: <task-queue-mgr>)
    ~empty?(a.task-queues);
end method;


// ====================================================
//    <task-queue-mgr>
// ====================================================
//    examples
// ====================================================

//  // =======================================================
//  // Globals
//  // =======================================================
//  define variable *scheduler* = make(<task-queue-mgr>, name: "My Scheduler");
//  define variable *scheduler-browser* = make(<scheduler-browser>, scheduler: *scheduler*);
//  
//  // ====================================================
//  // initialize the *scheduler*
//  // ====================================================
//  
//  create-queue(*scheduler*, name: "1 second", rate: make(<duration>, seconds: 1));
//  create-queue(*scheduler*, name: "5 second", rate: make(<duration>, seconds: 5));
//  create-queue(*scheduler*, name: "10 second", rate: make(<duration>, seconds: 10));
//  create-queue(*scheduler*, name: "1 minute", rate: make(<duration>, minutes: 1));
//
//  define method main () => ()
//  
//    let i = 0;
//    local method tick ()
//        if (i = 0)
//          format-out("tick\n");
//          i := 1;
//        else
//          format-out("tock\n");
//          i := 0;
//        end if;
//    end;
//  
//
//    // schedule the "task"
//    schedule(*scheduler*, name: "tick", task: tick, queue: "1 second");
//
//    // crank-up the scheduler (threads)
//    start-scheduler(*scheduler*);
//
//    // crank-up a browser so that we can see the scheduler details
//    make(<thread>, name: "scheduler browser",
//                   function: curry(start-frame, *scheduler-browser*));
//  
