Module:    TimeBase-Notification
Synopsis:  A soft realtime method of scheduling tasks.
Author:    Ron Franke-Polz
Copyright: (C) 2000, Ron Franke-Polz.  All rights reserved.
License:   This software is provided 'as-is', without any express or implied warranty. 
           In no event will the author be held liable for any damages arising from 
           the use of this software.
           Permission is granted to anyone to use this software for any purpose, 
           including commercial applications, and to alter it and redistribute it 
           freely, subject to the following restrictions:
           1. The origin of this software must not be misrepresented; you must not 
              claim that you wrote the original software. If you use this software in 
              a product, an acknowledgment in the product documentation would be appreciated 
              but is not required.
           2. Altered source versions must be plainly marked as such, and must not 
              be misrepresented as being the original software.
           3. This notice may not be removed or altered from any source distribution.

// ====================================================
//     <timebase-notification>
// ====================================================

define class <timebase-notification> (<object>)

  slot name :: <byte-string>,
    init-keyword: name:;

  slot rate :: <duration>,
    init-keyword: rate:;

  slot lock :: <lock>;

  slot notification :: <notification>;

  slot continue :: <boolean> = #f;          // scheduling status

end class;

// initialize
//
define method initialize (a :: <timebase-notification>, #keys)
    next-method();
    a.lock := make(<lock>);
    a.notification := make(<notification>, lock: a.lock);
end method;

// ====================================================
// example of <timebase-duration>
// ====================================================
//
//  let timebase1 = make(<timebase-notification>,
//                       name: "1 second timebase",
//                       rate: make(<duration>, seconds: 1));
//
//  keywords for <duration>
//
//  year:
//  month:
//  day:
//  hours:
//  minutes:
//  seconds:
//  microseconds:
//
// ====================================================


// start-notification-thread
//
// - start the scheduler in it's own thread.
// - the thread is returned.
//
define method start-notification-thread (a :: <timebase-notification>) => (thread :: <thread>)
    make(<thread>,
         name: a.name,
         function: method ()
                     execute(a);
                   end);
end method;


// execute
//
// - start the notification loop (loop forever)
// - we try to avoid having the interval slowly creep forward
//   in time by fixing a starting "now" time and adding the duration to it.
// - by re-figuring the "future" time to run by adding the "duration" to the
//   previous "future", we can accomodate any dither in execution time.
// - we just can't absolutely nail down the interval between executions,
//   but it's darn close.
//
define method execute (a :: <timebase-notification>)

    let now = current-date();
    let future = now + a.rate;

    // loop until we're told otherwise
    a.continue := #t;

    while (a.continue)
      with-lock(a.lock)
        // wake-up all of the threads that are sleeping waiting for this notification
        release-all(a.notification);
      end with-lock;

      // when we should next run is a fixed interval from the starting time
      future := future + a.rate;

      // what time is it now?
      now := current-date();

      // how long should we sleep until the next time to run
      let sleep-duration = future - now;
      let (sleep-days, sleep-hours, sleep-minutes, sleep-seconds, sleep-microseconds)
           = decode-duration(sleep-duration);

      // calculate the sleep (in seconds)
      let seconds :: <real> = sleep-days * 24.0;
      seconds := (seconds + sleep-hours) * 60.0;
      seconds := (seconds + sleep-minutes) * 60.0;
      seconds := seconds + sleep-seconds;
      seconds := seconds + (sleep-microseconds / 1000000.0);

      sleep(sleep-seconds);
    end while;

end method;



// with-timebase
//
// - a macro that wraps a body of code in a loop
//   that executes at a frequence specified in
//   a timebase (a <timebase-notification>).
//
define macro with-timebase
  { with-timebase (?var:expression) ?:body end }
  => {
       while (#t)
         with-lock(?var.lock)
           wait-for(?var.notification);
           ?body
         end with-lock;
       end while
     }
end macro;

// ====================================================
// example of with-timebase
// ====================================================
//
//  /* example #1 - within a thread */
//
//  let timebase1 = make(<timebase-notification>,
//                       name: "1 second timebase",
//                       rate: make(<duration>, seconds: 1));
//
//  start-notification-thread (timebase1);
//
//  with-timebase(timebase1)
//      format-out("tick (one second)\n");
//  end;
//
//
//  /* example #2 - execute a method in its own thread */
//
//  let timebase1 = make(<timebase-notification>,
//                       name: "1 second timebase",
//                       rate: make(<duration>, seconds: 1));
//
//  start-notification-thread (timebase1);
//
//  local method tick ()
//    with-timebase(timebase1)
//        format-out("tick (one second)\n");
//    end;
//  end;
//
//  let thread = make(<thread>, name: "tick",
//                             function: tick);
//



